from ..ioutils import read_int, read_option_item
from .linear_interpolation_variable import LinearInterpolationVariable


def read_variable(line):
    pos = read_int(line, "VARIABLE")

    if pos != 0:
        raise RuntimeError("Only one component is implemented for variables")

    vartype = read_option_item(line, "TYPE")[0]

    if vartype == "linearinterpolation":
        return LinearInterpolationVariable.read_variable(line)
    else:
        raise NotImplementedError(
            "The variable type {0} is not implemented yet!".format(vartype)
        )



from ..ioutils import line_option_list


class BaseVariable:
    def __init__(self, name):
        self.name = name

    def get_line(self):
        d = {"VARIABLE": 0, "NAME": self.name}
        return line_option_list(d)

    def write(self, dest):
        dest.write("{0}\n".format(self.get_line()))



#
# Maintainer: Amadeus Gebauer
#
# This is the interface between the meshio package and our Discretization description
# May not work perfectly
#
from typing import Any, Dict, List, Optional, Tuple

import meshio
import numpy as np

from .discretization import Discretization, Node
from .element.line2 import Line2
from .element.line3 import Line3
from .element.parse_element import create_element
from .element.quad4 import Quad4
from .element.tri3 import Tri3
from .element.tri6 import Tri6
from .nodeset import (
    LineNodeset,
    LineNodesetBuilder,
    PointNodeset,
    PointNodesetBuilder,
    SurfaceNodeset,
    SurfaceNodesetBuilder,
    VolumeNodeset,
    VolumeNodesetBuilder,
)
from .progress import progress

cell_nodes = {
    "vertex": 1,
    "line": 2,
    "line3": 3,
    "triangle": 3,
    "triangle6": 6,
    "tetra": 4,
    "tetra10": 10,
    "pyramid": 5,
    "hexahedron": 8,
    "hexahedron20": 20,
    "hexahedron27": 27,
    "wedge": 6,
    "quad": 4,
    "quad9": 9,
    "vertex": 1,
}

cell_disc_eles = {
    "vertex": "UNKNOWN",
    "line": "UNKNOWN",
    "line3": "UNKNOWN",
    "triangle": "UNKNOWN",
    "triangle6": "UNKNOWN",
    "tetra": "SOLIDT4",
    "tetra10": "SOLIDT10",
    "pyramid": "PYRAMID5",
    "hexahedron": "SOLIDH8",
    "hexahedron20": "SOLIDH20",
    "hexahedron27": "SOLIDH27",
    "wedge": "UNKNOWN",
    "quad": "UNKNOWN",
    "quad9": "UNKNOWN",
}

cell_disc_shape = {
    "vertex": "VERTEX1",
    "line": "LINE2",
    "line3": "LINE3",
    "triangle": "TRI3",
    "triangle6": "TRI6",
    "tetra": "TET4",
    "tetra10": "TET10",
    "pyramid": "PYRAMID5",
    "hexahedron": "HEX8",
    "hexahedron20": "HEX20",
    "hexahedron27": "HEX27",
    "wedge": "WEDGE6",
    "quad": "QUAD4",
    "quad9": "QUAD9",
}

_cell_data_id_names = ["medit:ref", "gmsh:geometrical"]

# invert cell_disc_shape
disc_shape_cell = {v: k for k, v in cell_disc_shape.items()}

cell_to_dim = {
    "vertex": 0,
    "line": 1,
    "line3": 1,
    "triangle": 2,
    "triangle6": 2,
    "tetra": 3,
    "tetra10": 3,
    "pyramid": 3,
    "hexahedron": 3,
    "hexahedron20": 3,
    "hexahedron27": 3,
    "wedge": 3,
    "quad": 2,
    "quad9": 2,
}

# meshio used vtk node ordering
# with this array, we reorder the nodes
ele_node_order_vtk2baci = {
    "VERTEX1": list(range(0, 1)),
    "LINE2": list(range(0, 2)),
    "LINE3": list(range(0, 2)),
    "TRI3": list(range(0, 3)),
    "TRI6": list(range(0, 6)),
    "TET4": list(range(0, 4)),
    "TET10": list(range(0, 10)),
    "PYRAMID5": list(range(0, 5)),
    "HEX8": list(range(0, 8)),
    "HEX20": [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        16,
        17,
        18,
        19,
        12,
        13,
        14,
        15,
    ],  # verify ?
    "HEX27": [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        16,
        17,
        18,
        19,
        12,
        13,
        14,
        15,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
    ],  # verify ?
    "WEDGE6": list(range(0, 6)),
    "QUAD4": list(range(0, 4)),
    "QUAD9": list(range(0, 9)),
}


def mesh2Discretization(mesh: meshio.Mesh) -> Discretization:
    # create empty discretization
    disc = Discretization()

    # mesh.points -> disc.nodes
    # mesh.cells -> disc.elements[Element.FieldTypeStructure]
    # node_sets -> are stored in disc.nodes

    # create Nodes from points
    for i, coord in progress(enumerate(mesh.points), label="Create nodes"):
        n = Node(coord)

        for key, v in mesh.point_data.items():
            n.data[key] = v[i]

        disc.nodes.append(n)

    # get the maximum element dimension, which is the dimension of the mesh
    maxdim = max([cell_to_dim[cell_block.type] for cell_block in mesh.cells])

    # create Elements from cells
    disc.elements.structure = []
    pointnsbuilder = PointNodesetBuilder()
    linensbuilder = LineNodesetBuilder()
    surfnsbuilder = SurfaceNodesetBuilder()
    volumensbuilder = VolumeNodesetBuilder()

    for cellgroupid, cellblock in enumerate(mesh.cells):
        if cellblock.type not in cell_nodes:
            raise Exception(
                "The cell type {0} is currently not implemented".format(cellblock.type)
            )

        for cellid, cell in enumerate(cellblock.data):
            # read nodes that belong to the cells
            nodes = []

            # in case of a vertex, cell is just an integer
            cell = np.array(cell).reshape((-1))

            for nodeids in cell:
                nodes.append(disc.nodes[nodeids])

            # cells with the same dimension as the mesh dimensions are normal cells
            # cells with a lower dimension are treated as surface, line or node
            # definitions
            element_dim = cell_to_dim[cellblock.type]
            if maxdim == element_dim:
                # this is a normal element
                ele = create_element(
                    cell_disc_eles[cellblock.type],
                    cell_disc_shape[cellblock.type],
                    # also need to reorder nodes in a few element types
                    [
                        nodes[i]
                        for i in ele_node_order_vtk2baci[
                            cell_disc_shape[cellblock.type]
                        ]
                    ],
                )
                disc.elements.structure.append(ele)

                # extract material info from cell data
                matid: Optional[int] = _get_material_from_cell_data(
                    mesh.cell_data, cellgroupid, cellid
                )
                if matid == None:
                    matid = 1
                ele.options["MAT"] = matid

                # extract cell data
                for key, data in mesh.cell_data.items():
                    ele.data[key] = data[cellgroupid][cellid]

                # create surface-nodesets
                if maxdim == 2 and len(mesh.cell_data) > 0:
                    nsid: Optional[int] = _get_nodesetid_from_cell_data(
                        mesh.cell_data, cellgroupid, cellid
                    )

                    if nsid is not None:
                        for node in nodes:
                            surfnsbuilder.add(node, nsid)

            else:
                # this is a lower-dimensional element
                # treat as surface-, line- or point-nodeset definition

                # extract nodeset info from cell data
                nsid: Optional[int] = _get_nodesetid_from_cell_data(
                    mesh.cell_data, cellgroupid, cellid
                )

                if nsid is not None:
                    for node in nodes:
                        if element_dim == 0:
                            pointnsbuilder.add(node, nsid)
                        elif element_dim == 1:
                            linensbuilder.add(node, nsid)
                        elif element_dim == 2:
                            surfnsbuilder.add(node, nsid)

    # go through nodesets
    for name, nodes in mesh.point_sets.items():
        if "volume" in name:
            nsid = volumensbuilder.get_unused_id()
            for n in nodes:
                volumensbuilder.add(disc.nodes[n], nsid)
        if "surface" in name:
            nsid = surfnsbuilder.get_unused_id()
            for n in nodes:
                surfnsbuilder.add(disc.nodes[n], nsid)
        elif "line" in name:
            nsid = linensbuilder.get_unused_id()
            for n in nodes:
                linensbuilder.add(disc.nodes[n], nsid)
        elif "point" in name:
            nsid = pointnsbuilder.get_unused_id()
            for n in nodes:
                pointnsbuilder.add(disc.nodes[n], nsid)

    disc.volumenodesets = volumensbuilder.finalize()
    disc.pointnodesets = pointnsbuilder.finalize()
    disc.linenodesets = linensbuilder.finalize()
    disc.surfacenodesets = surfnsbuilder.finalize()

    disc.finalize()
    return disc


def discretization2mesh(dis: Discretization) -> meshio.Mesh:
    dis.compute_ids(zero_based=True)

    points = dis.get_node_coords()

    point_data = {}

    for i, n in enumerate(dis.nodes):
        for k, v in n.data.items():
            if k not in point_data:
                point_data[k] = np.zeros(
                    tuple([len(dis.nodes)] + list(np.array(v).shape))
                )

            point_data[k][i] = v

    cells = []
    cell_data = {}

    id_to_index = {}

    # store cells
    for eletype in dis.elements.values():
        for ele in eletype:
            celltype = disc_shape_cell[ele.shape]

            newgroup = False

            if len(cells) == 0 or cells[-1][0] != celltype:
                cell_nodes = []
                cells.append((celltype, cell_nodes))
                newgroup = True
            else:
                cell_nodes = cells[-1][1]

            cell_nodes.append([n.id for n in ele.nodes])
            id_to_index[ele.id] = (len(cells) - 1, len(cell_nodes) - 1)

    # go again through the elements to store element data
    for eletype in dis.elements.values():
        for ele in eletype:
            block_id, cell_id = id_to_index[ele.id]

            if "MAT" in ele.options:
                if "material" not in cell_data:
                    cell_data["material"] = _create_cell_variable(
                        cells, tuple([1]), int
                    )

                cell_data["material"][block_id][cell_id] = int(
                    ele.options["MAT"][0]
                    if _isiter(ele.options["MAT"])
                    else ele.options["MAT"]
                )

            for variable_name, value in ele.data.items():
                value_reshaped = np.array(value).reshape((-1))

                if variable_name not in cell_data:
                    cell_data[variable_name] = _create_cell_variable(
                        cells, value_reshaped.shape, value_reshaped.dtype
                    )
                    cell_data["material"][block_id][cell_id] = value_reshaped

    mesh: meshio.Mesh = meshio.Mesh(
        points, cells, cell_data=cell_data, point_data=point_data
    )

    dis.reset()

    return mesh


def _create_cell_variable(
    cells: List[Tuple[str, List[int]]], shape: Tuple[int], dtype
) -> List[np.ndarray]:
    data = []

    for block in cells:
        new_shape = tuple([len(block[1])] + list(shape))
        if len(shape) == 1 and shape[0] == 1:
            new_shape = tuple([len(block[1])])
        data.append(np.zeros(new_shape, dtype=dtype))

    return data


def _get_material_from_cell_data(
    celldata: Dict[str, List[List[Any]]], cellgroupid: int, cellid: int
) -> Optional[int]:
    for name in _cell_data_id_names:
        if name in celldata:
            return int(celldata[name][cellgroupid][cellid])

    return None


def _get_nodesetid_from_cell_data(
    celldata: Dict[str, List[List[Any]]], cellgroupid: int, cellid: int
) -> Optional[int]:
    for name in _cell_data_id_names:
        if name in celldata:
            return int(celldata[name][cellgroupid][cellid])

    return None


def _isiter(list: Any) -> bool:
    try:
        iter(list)
    except TypeError:
        return False

    return True



from collections import OrderedDict

from ..ioutils import read_option_item, write_title
from .component import Component
from .variable import BaseVariable
from .variablereader import read_variable


class Function:
    def __init__(self, id):
        self.id = id
        self.components = []
        self.variables = {}

    def reset(self):
        self.id = None

    def add_component(self, component: Component):
        if component.pos < len(self.components):
            raise RuntimeError("This component is already added to the function")
        elif component.pos > len(self.components):
            raise RuntimeError(
                "There is at least one component missing. Got component {0} but expected {1}".format(
                    component.pos, len(self.components)
                )
            )

        self.components.append(component)

    def add_variable(self, variable: BaseVariable):
        if variable.name in self.variables:
            raise RuntimeError("The variable with the same name already exists")

        self.variables[variable.name] = variable

    def get_sections(self):
        d = OrderedDict()

        lines = []

        for c in self.components:
            lines.append(c.get_line())

        for v in self.variables.values():
            lines.append(v.get_line())
        d["FUNCT{0}".format(self.id)] = lines

        return d

    def write(self, dest):
        write_title(dest, "FUNCT{0}".format(self.id), True)

        for c in self.components:
            c.write(dest)

        for v in self.variables.values():
            v.write(dest)

    @staticmethod
    def read_functions(sections):
        # read functions
        maxf = 10

        functions = []
        for i in range(maxf):
            if "FUNCT{0}".format(i) in sections:
                # function exists
                # parse function

                curr_function = None
                # read each component
                for line in sections["FUNCT{0}".format(i)]:
                    component = None
                    variable = None
                    if read_option_item(line, "COMPONENT")[0] is not None:
                        # this is a component
                        component = Component.read_component(line)
                    elif read_option_item(line, "VARIABLE")[0] is not None:
                        # this is a variable
                        variable = read_variable(line)

                    if component is not None:
                        # add to function
                        if curr_function is None:
                            curr_function = Function(i)
                            if i - 1 != len(functions):
                                raise RuntimeError(
                                    "The functions are not in the correct order"
                                )

                        curr_function.add_component(component)

                    if variable is not None:
                        curr_function.add_variable(variable)

                if curr_function is not None:
                    functions.append(curr_function)

        return functions



from ..ioutils import (
    line_option_list,
    read_floats,
    read_int,
    read_ints,
    read_option_item,
    write_option_list,
)
from .variable import BaseVariable


class LinearInterpolationVariable(BaseVariable):
    def __init__(self, name: str, times, values):
        super(LinearInterpolationVariable, self).__init__(name)
        self.times = times
        self.values = values

    def get_line(self):
        line = super(LinearInterpolationVariable, self).get_line()

        d = {
            "TYPE": "linearinterpolation",
            "NUMPOINTS": len(self.times),
            "TIMES": self.times,
            "VALUES": self.values,
        }
        return "{0} {1}".format(line, line_option_list(d))

    def write(self, dest):
        dest.write("{0}\n".format(self.get_line()))

    @staticmethod
    def read_variable(line):
        numpoints = read_int(line, "NUMPOINTS")

        times = read_floats(line, "TIMES", numpoints)
        values = read_floats(line, "VALUES", numpoints)

        name = read_option_item(line, "NAME")[0]
        if name is None:
            raise RuntimeError(
                "Expecting NAME of variable to be given. Line is {0}".format(line)
            )

        return LinearInterpolationVariable(name, times, values)



import io
import os
import unittest

import lnmmeshio
import numpy as np
from lnmmeshio import ioutils
from lnmmeshio.nodeset import LineNodeset, PointNodeset, SurfaceNodeset, VolumeNodeset

script_dir = os.path.dirname(os.path.realpath(__file__))


class TestDiscretizationIO(unittest.TestCase):
    def setUp(self):
        pass

    def test_read(self):
        # read dat file
        mesh = lnmmeshio.read_mesh(os.path.join(script_dir, "data", "dummy.dat"))

        # check, whether all nodes are read correctly
        self.assertEqual(mesh.points.shape, (224, 3))

        # check, whether all tet elements were read correctly
        self.assertEqual(mesh.get_cells_type("tetra").shape, (12, 4))
        self.assertEqual(mesh.get_cells_type("tetra10").shape, (12, 10))
        self.assertEqual(mesh.cell_data["material"][0][0], 1)
        self.assertEqual(mesh.cell_data["material"][1][0], 4)
        self.assertEqual(mesh.cell_data["material"][2][0], 5)
        self.assertEqual(mesh.cell_data["material"][3][0], 6)

        # check, whether all hex elements were read correctly
        self.assertEqual(mesh.get_cells_type("hexahedron").shape, (65, 8))

    def test_write(self):
        # read dat file
        mesh = lnmmeshio.read(os.path.join(script_dir, "data", "dummy.dat"), out=True)

        if not os.path.isdir(os.path.join(script_dir, "tmp")):
            os.makedirs(os.path.join(script_dir, "tmp"))

        # write same dat file
        lnmmeshio.write(os.path.join(script_dir, "tmp", "gen.dat"), mesh)

        # check, whether both files are identical
        # self.assertTrue(filecmp.cmp(os.path.join(script_dir, 'data', 'dummy.dat'),
        #    os.path.join((script_dir), 'tmp', 'gen.dat')))

        # check, whether both sections do coincide
        with open(os.path.join(script_dir, "data", "dummy.dat"), "r") as f:
            sections1 = ioutils.read_dat_sections(f)
        with open(os.path.join(script_dir, "tmp", "gen.dat"), "r") as f:
            sections2 = ioutils.read_dat_sections(f)

        self.assertListEqual(
            sorted(list(sections1.keys())), sorted(list(sections2.keys()))
        )

        for key in sections1.keys():
            self.assertListEqual(sorted(sections1[key]), sorted(sections2[key]))

    def test_read_new(self):
        # read discretization
        disc = lnmmeshio.read(os.path.join(script_dir, "data", "dummy.dat"), out=True)

        disc.compute_ids(zero_based=False)

        self.assertEqual(len(disc.nodes), 224)
        self.assertEqual(len(disc.elements.structure), 89)

        self.assertListEqual([ns.id for ns in disc.nodes[0].surfacenodesets], [1, 13])
        self.assertListEqual([ns.id for ns in disc.nodes[47].surfacenodesets], [9])
        self.assertListEqual([ns.id for ns in disc.nodes[147].surfacenodesets], [6])

    def test_write_new(self):
        # build dummy discretization
        d: lnmmeshio.Discretization = lnmmeshio.Discretization()
        d.nodes = [
            lnmmeshio.Node(np.array([0.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([1.0, 0.0, 0.0])),
            lnmmeshio.Node(np.array([0.0, 1.0, 0.0])),
            lnmmeshio.Node(np.array([0.0, 0.0, 1.0])),
        ]
        for i in range(0, 4):
            d.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber1] = lnmmeshio.Fiber(
                np.array([1.0, 0.0, 0.0])
            )
            d.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber2] = lnmmeshio.Fiber(
                np.array([0.0, 1.0, 0.0])
            )

        d.pointnodesets.append(PointNodeset(1))
        d.linenodesets.append(LineNodeset(1))
        d.surfacenodesets.append(SurfaceNodeset(1))
        d.volumenodesets.append(VolumeNodeset(1))

        d.pointnodesets[0].add_node(d.nodes[0])

        # dline
        for i in range(0, 2):
            d.linenodesets[0].add_node(d.nodes[i])

        # dsurf
        for i in range(0, 3):
            d.surfacenodesets[0].add_node(d.nodes[i])

        # dvol
        for i in range(0, 4):
            d.volumenodesets[0].add_node(d.nodes[i])

        d.elements.structure = [lnmmeshio.Element("SOLIDT4SCATRA", "TET4", d.nodes)]
        d.elements.structure[0].options = {
            "MAT": 1,
            "KINEM": "nonlinear",
            "TYPE": "Std",
        }

        dummy_file = io.StringIO()
        d.write(dummy_file)
        dummy_file.seek(0)

        # read dummy file
        d_new = lnmmeshio.read_baci_discr(dummy_file)
        d_new.compute_ids(zero_based=True)

        self.assertEqual(len(d_new.nodes), 4)
        self.assertEqual(len(d_new.elements.structure), 1)

        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[0].coords - np.array([0, 0, 0])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[1].coords - np.array([1, 0, 0])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[2].coords - np.array([0, 1, 0])), 0.0
        )
        self.assertAlmostEqual(
            np.linalg.norm(d_new.nodes[3].coords - np.array([0, 0, 1])), 0.0
        )

        for i in range(0, 4):
            self.assertAlmostEqual(
                np.linalg.norm(
                    d_new.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber1].fiber
                    - np.array([1, 0, 0])
                ),
                0.0,
            )
            self.assertAlmostEqual(
                np.linalg.norm(
                    d_new.nodes[i].fibers[lnmmeshio.Fiber.TypeFiber2].fiber
                    - np.array([0, 1, 0])
                ),
                0.0,
            )

        self.assertListEqual([n.id for n in d_new.pointnodesets], [0])
        self.assertListEqual(
            [sorted([n.id for n in nds]) for nds in d_new.linenodesets], [[0, 1]]
        )
        self.assertListEqual(
            [sorted([n.id for n in nds]) for nds in d_new.surfacenodesets], [[0, 1, 2]]
        )
        self.assertListEqual(
            [sorted([n.id for n in nds]) for nds in d_new.volumenodesets],
            [[0, 1, 2, 3]],
        )

        self.assertEqual(d_new.elements.structure[0].type, "SOLIDT4SCATRA")
        self.assertEqual(d_new.elements.structure[0].shape, "TET4")
        self.assertListEqual(d_new.elements.structure[0].nodes, d_new.nodes)
        self.assertEqual(d_new.elements.structure[0].options["MAT"], "1")
        self.assertEqual(d_new.elements.structure[0].options["KINEM"], "nonlinear")
        self.assertEqual(d_new.elements.structure[0].options["TYPE"], "Std")

    def test_io_utils_text_fill(self):
        # test text_fill
        self.assertEqual(
            lnmmeshio.ioutils.text_fill("TEST", 10, "-", minimum=10, fill_left=True),
            "----------TEST",
        )
        self.assertEqual(
            lnmmeshio.ioutils.text_fill("TEST", 10, "-", minimum=3, fill_left=True),
            "------TEST",
        )
        self.assertEqual(
            lnmmeshio.ioutils.text_fill("TEST2", 12, "-", minimum=3), "TEST2-------"
        )

    def test_io_utils_write_title(self):
        dummy_file = io.StringIO()

        # test write_title
        lnmmeshio.ioutils.write_title(dummy_file, "TITLE")
        self.assertEqual(
            dummy_file.getvalue(),
            "--------------------------------------------------------------------TITLE\n",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_title(
            dummy_file,
            "THIS IS A VERY LONG TITLE WHICH TAKES MORE SPACE THAN ALLOWED LETS SEE HOW",
        )
        self.assertEqual(
            dummy_file.getvalue(),
            "---THIS IS A VERY LONG TITLE WHICH TAKES MORE SPACE THAN ALLOWED LETS SEE HOW\n",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_title(dummy_file, "TITLE", newline=False)
        self.assertEqual(
            dummy_file.getvalue(),
            "--------------------------------------------------------------------TITLE",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_write_option(self):
        dummy_file = io.StringIO()
        # test write_option
        lnmmeshio.ioutils.write_option(dummy_file, "KEY", "VALUE")
        self.assertEqual(
            dummy_file.getvalue(), "KEY                             VALUE\n"
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_option(dummy_file, "KEY", "VALUE", comment="COMMENT")
        self.assertEqual(
            dummy_file.getvalue(),
            "KEY                             VALUE                           // COMMENT\n",
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_option(dummy_file, "KEY", "VALUE", newline=False)
        self.assertEqual(dummy_file.getvalue(), "KEY                             VALUE")
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_write_option_list(self):
        dummy_file = io.StringIO()
        # test write_option_list
        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {"KEY0": "VAL0", "KEY1": ["VAL1.1", "VAL1.2"], "KEY2": [2.1, 2.2]},
        )
        self.assertEqual(
            dummy_file.getvalue(), "KEY0 VAL0 KEY1 VAL1.1 VAL1.2 KEY2 2.1 2.2\n"
        )
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {"KEY1": ["VAL1.1", "VAL1.2"], "KEY2": [2.1, 2.2]},
            newline=False,
        )
        self.assertEqual(dummy_file.getvalue(), "KEY1 VAL1.1 VAL1.2 KEY2 2.1 2.2")
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_write_comment(self):
        dummy_file = io.StringIO()
        # test write_comment
        lnmmeshio.ioutils.write_comment(dummy_file, "COMMENT")
        self.assertEqual(dummy_file.getvalue(), "// COMMENT\n")
        dummy_file.truncate(0)
        dummy_file.seek(0)

        lnmmeshio.ioutils.write_comment(dummy_file, "COMMENT", newline=False)
        self.assertEqual(dummy_file.getvalue(), "// COMMENT")
        dummy_file.truncate(0)
        dummy_file.seek(0)

    def test_io_utils_read_dat_sections(self):
        # build file
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_title(dummy_file, "HEAD1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY1", "VAL1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY2", "VAL2")
        lnmmeshio.ioutils.write_title(dummy_file, "HEAD2")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY1", "VAL1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY2", "VAL2")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY3", "VAL3")
        lnmmeshio.ioutils.write_title(dummy_file, "HEAD3")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY1", "VAL1")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY2", "VAL2")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY3", "VAL3")
        lnmmeshio.ioutils.write_option(dummy_file, "KEY4", "VAL4")

        # seek to beginning
        dummy_file.seek(0)
        sections = lnmmeshio.ioutils.read_dat_sections(dummy_file)
        self.assertListEqual(list(sections.keys()), ["", "HEAD1", "HEAD2", "HEAD3"])
        self.assertListEqual(sections[""], [])
        self.assertEqual(len(sections["HEAD1"]), 2)
        self.assertEqual(len(sections["HEAD2"]), 3)
        self.assertEqual(len(sections["HEAD3"]), 4)

    def test_io_utils_read_option_item(self):
        # build file
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {
                "KEY1": "VAL1",
                "KEY2": ["VAL2.1", "VAL2.2"],
                "KEY3": 3,
                "KEY4": [4.1, 4.2],
            },
        )

        line = dummy_file.getvalue()
        self.assertEqual(lnmmeshio.ioutils.read_option_item(line, "KEY1")[0], "VAL1")
        self.assertListEqual(
            lnmmeshio.ioutils.read_option_items(line, "KEY2", num=2)[0],
            ["VAL2.1", "VAL2.2"],
        )
        self.assertEqual(lnmmeshio.ioutils.read_option_item(line, "KEY3")[0], "3")
        self.assertListEqual(
            lnmmeshio.ioutils.read_option_items(line, "KEY4", num=2)[0], ["4.1", "4.2"]
        )

        self.assertRaises(
            RuntimeError, lambda: lnmmeshio.ioutils.read_option_item(line, "KEY5")
        )

    def test_io_utils_read_next_option(self):
        # build file
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_option_list(
            dummy_file,
            {
                "KEY1": "VAL1",
                "KEY2": ["VAL2.1", "VAL2.2"],
                "KEY3": 3,
                "KEY4": [4.1, 4.2],
            },
        )

        line = dummy_file.getvalue()

        for i, (key, values) in enumerate(
            lnmmeshio.ioutils.read_key_values(
                line, lambda key: 2 if key == "KEY2" or key == "KEY4" else 1
            )
        ):
            self.assertEqual(key, f"KEY{i+1}")

            if i == 0:
                self.assertEqual(values, ["VAL1"])
            elif i == 1:
                self.assertListEqual(values, ["VAL2.1", "VAL2.2"])
            elif i == 2:
                self.assertListEqual(values, ["3"])
            elif i == 3:
                self.assertListEqual(values, ["4.1", "4.2"])

    def test_dummy_read(self):
        dummy_file = io.StringIO()
        lnmmeshio.ioutils.write_title(dummy_file, "TITLE", newline=True)
        lnmmeshio.ioutils.write_title(dummy_file, "TITLE", newline=True)
        dummy_file.seek(0)

        with self.assertRaises(ValueError) as _:
            lnmmeshio.ioutils.read_dat_sections(dummy_file)


if __name__ == "__main__":
    unittest.main()



from ..ioutils import line_option_list, read_int, read_option_item


class Component:
    def __init__(self, expression, pos):
        self.expression = expression
        self.pos = pos

    def write(self, dest):
        dest.write("{0}\n".format(self.get_line()))

    def get_line(self):
        d = {"COMPONENT": self.pos, "FUNCTION": self.expression}
        return line_option_list(d)

    @staticmethod
    def read_component(line):
        pos = read_int(line, "COMPONENT")

        if pos is None:
            return None

        expression = read_option_item(line, "FUNCTION")[0]
        component = Component(expression, pos)

        return component



import time
from typing import Iterable, Iterator, Sized, Union, overload

from progress import bar, spinner


def progress(
    iterator: Union[Iterable, Sized],
    out: bool = True,
    label: str = None,
    btype: str = "bar",
) -> Union[Iterable, Sized]:
    max_val = 0
    try:
        max_val = len(iterator)
    except TypeError:
        max_val = -1

    if max_val == 0:
        out = False

    if out:
        if label is None:
            label = "progress"

        if max_val < 0:
            return ProgressIterator(FastSpinner(label).iter(iterator), label)
        else:
            return ProgressIterator(FastBar(label).iter(iterator), label)
    else:
        return iterator


def clearln():
    print("\033[F\r\x1b[K", end="")


class FastBar(bar.Bar):
    suffix = "%(percent).1f%% - %(eta_td)s"

    def __init__(self, label):
        super(FastBar, self).__init__(label)
        self.last = time.time()

    def update(self):
        if time.time() - self.last > 0.1:
            self.last = time.time()
            super(FastBar, self).update()


class FastSpinner(spinner.Spinner):
    def __init__(self, label):
        super(FastSpinner, self).__init__(label)
        self.last = time.time()

    def update(self):
        if time.time() - self.last > 0.1:
            self.last = time.time()
            super(FastSpinner, self).update()


class ProgressIterator:
    def __init__(self, iterator, label):
        self.iterator = iterator
        self.starttime = time.time()
        self.label = label

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.iterator.__next__()
        except StopIteration:
            # print total time
            clearln()
            print("{0} ... {1:.2f}s".format(self.label, time.time() - self.starttime))
            raise StopIteration
